﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using ICSharpCode.AvalonEdit;

namespace AB3.AppDesigner.Controls
{
	/// <summary>
	/// 
	/// </summary>
	/// <see cref="http://www.codeproject.com/Articles/84213/How-to-add-a-Close-button-to-a-WPF-TabItem"/>
	public class CloseableTabItem: TabItem
	{
		private bool _InSetContextText = false;
		private string _Title = "TabItem";

		public CloseableTabItem()
		{
			// Create an instance of the usercontrol
			CloseableHeader closableTabHeader = new CloseableHeader();

			// Assign the usercontrol to the tab header
			this.Header = closableTabHeader;

			// Attach to the CloseableHeader events (Mouse Enter/Leave, Button Click, and Label resize)
			closableTabHeader.button_close.MouseEnter += new MouseEventHandler(button_close_MouseEnter);
			closableTabHeader.button_close.MouseLeave += new MouseEventHandler(button_close_MouseLeave);
			closableTabHeader.button_close.Click += new RoutedEventHandler(button_close_Click);
			closableTabHeader.label_TabTitle.SizeChanged += new SizeChangedEventHandler(label_TabTitle_SizeChanged);
						
			this.ContentEditor = new TextEditor();
			this.ContentEditor.HorizontalAlignment = System.Windows.HorizontalAlignment.Stretch;
			this.ContentEditor.VerticalAlignment = System.Windows.VerticalAlignment.Stretch;
			this.ContentEditor.ShowLineNumbers = true;
			this.ContentEditor.Text = "My Text";
			this.Content = this.ContentEditor;
			this.ContentEditor.TextChanged += ContentEditor_TextChanged;
		}

		private void ContentEditor_TextChanged(object sender, EventArgs e)
		{
			if (!this._InSetContextText)
			{
				this.ContentTextChanged = true;
				this.SetTitle();
			}
		}

		public bool ContentTextChanged { get; private set; }

		/// <summary>
		/// Property - Set the Title of the Tab
		/// </summary>
		public string Title
		{
			get
			{
				return this._Title;
			}
			set
			{
				this._Title = value;
				this.SetTitle();
			}
		}

		private void SetTitle()
		{
			CloseableHeader tHeader = this.Header as CloseableHeader;
			if (tHeader != null)
			{
				tHeader.label_TabTitle.Content = this._Title;
				if (this.ContentTextChanged)
				{
					tHeader.label_TabTitle.Content = this._Title + " (Changed)";
					tHeader.label_TabTitle.FontStyle = System.Windows.FontStyles.Italic;
					tHeader.label_TabTitle.FontWeight = System.Windows.FontWeights.Bold;
				}
				else
				{
					tHeader.label_TabTitle.Content = this._Title;
					tHeader.label_TabTitle.FontStyle = System.Windows.FontStyles.Normal;
					tHeader.label_TabTitle.FontWeight = System.Windows.FontWeights.Normal;
				}
			}
		}

		public string ContentText
		{
			get { return this.ContentEditor.Text; }
			set 
			{
				this._InSetContextText = true;
				this.ContentEditor.Text = value;
				this._InSetContextText = false;
			}
		}


		public TextEditor ContentEditor { get; private set; }

		// Override OnSelected - Show the Close Button
		protected override void OnSelected(RoutedEventArgs e)
		{
			base.OnSelected(e);
			((CloseableHeader)this.Header).button_close.Visibility = Visibility.Visible;
		}

		// Override OnUnSelected - Hide the Close Button
		protected override void OnUnselected(RoutedEventArgs e)
		{
			base.OnUnselected(e);
			((CloseableHeader)this.Header).button_close.Visibility = Visibility.Hidden;
		}

		// Override OnMouseEnter - Show the Close Button
		protected override void OnMouseEnter(MouseEventArgs e)
		{
			base.OnMouseEnter(e);
			((CloseableHeader)this.Header).button_close.Visibility = Visibility.Visible;
		}

		// Override OnMouseLeave - Hide the Close Button (If it is NOT selected)
		protected override void OnMouseLeave(MouseEventArgs e)
		{
			base.OnMouseLeave(e);
			if (!this.IsSelected)
			{
				((CloseableHeader)this.Header).button_close.Visibility = Visibility.Hidden;
			}
		}

		// Button MouseEnter - When the mouse is over the button - change color to Red
		private void button_close_MouseEnter(object sender, MouseEventArgs e)
		{
			((CloseableHeader)this.Header).button_close.Foreground = Brushes.Red;
		}

		// Button MouseLeave - When mouse is no longer over button - change color back to black
		private void button_close_MouseLeave(object sender, MouseEventArgs e)
		{
			((CloseableHeader)this.Header).button_close.Foreground = Brushes.Black;
		}


		// Button Close Click - Remove the Tab - (or raise an event indicating a "CloseTab" event has occurred)
		private void button_close_Click(object sender, RoutedEventArgs e)
		{
			((TabControl)this.Parent).Items.Remove(this);
		}


		// Label SizeChanged - When the Size of the Label changes (due to setting the Title) set position of button properly
		private void label_TabTitle_SizeChanged(object sender, SizeChangedEventArgs e)
		{
			((CloseableHeader)this.Header).button_close.Margin = new Thickness(((CloseableHeader)this.Header).label_TabTitle.ActualWidth + 5, 3, 4, 0);
		}
	}
}
