﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using AB3.DataTypes;

namespace AB3.Helpers
{
	/// <summary>Helper to convert from a FDL line to a FDL datatype
	/// </summary>
	public static class LineConverter
	{
		/// <summary>Converts a FDL line to the name of the FDLFlowDefinition
		/// (Keyword = 'FlowDefinition:')
		/// </summary>
		/// <param name="line">Line of text to parse</param>
		/// <returns>Name of the FDLFlowDefinition</returns>
		public static string ToFlowDefinitionName(string line)
		{
			string tName = line.Substring(15).Replace("{", string.Empty).Trim();

			return tName;
		}

		/// <summary>Converts a FDL line to the name of the FDLRemarkDefinition
		/// (Keyword = 'Remark:')
		/// </summary>
		/// <param name="line">Line of text to parse</param>
		/// <returns>Name of the FDLRemarkDefinition</returns>
		public static string ToRemarkDefinitionName(string line)
		{
			string tName = line.Substring(7).Replace("{", string.Empty).Trim();

			return tName;
		}

		/// <summary>Converts a FDL line to a FDLStepDefinition
		/// (Keywords = 'Step:' or 'Flow:'
		/// </summary>
		/// <param name="app">App definition</param>
		/// <param name="lineNumber">Line number of text to parse</param>
		/// <returns>Instance of FDLStepDefinition</returns>
		public static FDLStepDefinition ToStepDefinition(AppDefinition app, int lineNumber)
		{
			string tLine = app.FlowText[lineNumber];
			FDLStepDefinition tStep = new FDLStepDefinition();
			tStep.LineNumber = lineNumber;
			if (tLine.StartsWith("Flow:"))
				tStep.IsSubFlowStep = true;

			tLine = tLine.Substring(5);
			string[] tLineParts = tLine.Split(new string[] { "=>" }, StringSplitOptions.None);

			tStep.TypeName = tLineParts[0].Replace("{", string.Empty).Trim();
			if (tLineParts.Length != 1)
			{
				tStep.CustomName = tLineParts[1].Replace("{", string.Empty).Trim();
				if (string.IsNullOrEmpty(tStep.CustomName))
					tStep.CustomName = tStep.TypeName;
			}
			else
				tStep.CustomName = tStep.TypeName;

			return tStep;
		}

		private static Regex _AttributeExpresseion = new Regex(@"(\w+)(\s*)(=)(\s*)("")([^""]*)("")", RegexOptions.Singleline);

		/// <summary>Converts a FDL line to multiple FDLDesignerAttributes
		/// </summary>
		/// <param name="app">App definition</param>
		/// <param name="lineNumber">Line number of text to parse</param>
		/// <returns>Instances of FDLDesignerAttribute</returns>
		public static IEnumerable<FDLAttribute> ToAttributes(AppDefinition app, int lineNumber)
		{
			//See: http://stackoverflow.com/questions/4529608/combine-two-regex-group-into-a-key-value-pair-object
			//Regex tExpression = new Regex(@"(\w+)(\s*)(=)(\s*)("")([^""]*)("")", RegexOptions.Singleline); // | RegexOptions.Compiled);
			List<FDLAttribute> tAttributes = new List<FDLAttribute>();
			FDLAttribute tAttribute;
			foreach (Match tMatch in _AttributeExpresseion.Matches(app.FlowText[lineNumber]))
			{
				tAttribute = new FDLAttribute();
				tAttribute.LineNumber = lineNumber;
				tAttribute.Name = tMatch.Groups[1].Value;
				tAttribute.Value = tMatch.Groups[6].Value;
				tAttributes.Add(tAttribute);
			}

			return tAttributes;
		}

		/// <summary>Converts a FDL line to a FDLDesignerAttribute
		/// </summary>
		/// <param name="app">App definition</param>
		/// <param name="lineNumber">Line number of text to parse</param>
		/// <returns>Instance of FDLDesignerAttribute</returns>
		public static FDLAttribute ToAttribute(AppDefinition app, int lineNumber)
		{
			string tLine = app.FlowText[lineNumber];
			if (tLine.EndsWith("{"))
				return null;
			else if (string.IsNullOrEmpty(tLine))
				return null;
			else if (tLine == "{")
				return null;
			else if (tLine == "}")
				return null;
			else if (!tLine.Contains('='))
			{
				FDLException tEx = new FDLException("Syntax error in attribute: '=' is missing!",
					lineNumber + 1, app.FlowText[lineNumber]);
				app.ErrorLogger.LogError(tEx.ToString());
				return null;
			}
			else
			{
				string[] tLineParts = app.FlowText[lineNumber].Split(new string[] { "=" }, StringSplitOptions.None);
				FDLAttribute tAttribute = new FDLAttribute();
				tAttribute.Name = tLineParts[0].Replace("{", string.Empty).Trim();
				tAttribute.Value = tLineParts[1].Trim();
				tAttribute.LineNumber = lineNumber;

				int tPosition = tAttribute.Value.IndexOf('"');
				if (tPosition != -1)
				{
					tAttribute.Value = tAttribute.Value.Remove(0, tPosition + 1);
					tPosition = tAttribute.Value.IndexOf('"');
					if (tPosition != -1)
					{
						tAttribute.Value = tAttribute.Value.Substring(0, tPosition);
						return tAttribute;
					}
					else
					{
						FDLException tEx = new FDLException("Syntax error in attribute: '\"' is missing!",
							lineNumber + 1, app.FlowText[lineNumber]);
						app.ErrorLogger.LogError(tEx.ToString());
						return null;
					}
				}
				else
				{
					FDLException tEx = new FDLException("Syntax error in attribute: '\"' is missing!",
						lineNumber + 1, app.FlowText[lineNumber]);
					app.ErrorLogger.LogError(tEx.ToString());
					return null;
				}
			}
		}


		/// <summary>Converts a FDL line to a FDLInterfaceDefinition
		/// </summary>
		/// <param name="app">App definition</param>
		/// <param name="lineNumber">Line number of text to parse</param>
		/// <returns>Instance of FDLInterfaceDefinition</returns>
		public static FDLInterfaceDefinition ToInterfaceDefinition(AppDefinition app, int lineNumber)
		{
			FDLInterfaceDefinition tInterface = new FDLInterfaceDefinition();
			tInterface.LineNumber = lineNumber;

			if (app.FlowText[lineNumber].StartsWith("In:"))
			{
				tInterface.Direction = Directions.In;
				tInterface.Name = app.FlowText[lineNumber].Substring(3).Replace("{", string.Empty).Trim();
			}
			else
			{
				tInterface.Direction = Directions.Out;
				tInterface.Name = app.FlowText[lineNumber].Substring(4).Replace("{", string.Empty).Trim();
				
			}

			return tInterface;
		}

		/// <summary>Converts a FDL line to a FDLWireDefinition
		/// </summary>
		/// <param name="app">App definition</param>
		/// <param name="lineNumber">Line number of text to parse</param>
		/// <returns>Instance of FDLWireDefinition</returns>
		public static FDLWireDefinition ToWireDefinition(AppDefinition app, int lineNumber)
		{
			string[] tLineParts = app.FlowText[lineNumber].Substring(5).Trim().Split(new string[] { "->" }, StringSplitOptions.None);
			FDLWireDefinition tWireDefinition = new FDLWireDefinition();
			tWireDefinition.LineNumber = lineNumber;

			if (tLineParts.Length == 1)
			{
				FDLException tEx = new FDLException("Syntax error in wire definition!", lineNumber + 1, app.FlowText[lineNumber]);
				app.ParseExceptions.Add(tEx);
				app.ErrorLogger.LogError(tEx.ToString());
				tWireDefinition.Source.Step = "Unknown";
				tWireDefinition.Source.Pin = "Unknown";
				tWireDefinition.Target.Step = "Unknown";
				tWireDefinition.Target.Pin = "Unknown";
			}
			else
			{
				string[] tSourceParts = tLineParts[0].Split(new string[] { "." }, StringSplitOptions.None);
				if (tSourceParts.Length == 1)
				{
					FDLException tEx = new FDLException("Syntax error in wire source!", lineNumber + 1, app.FlowText[lineNumber]);
					app.ParseExceptions.Add(tEx);
					app.ErrorLogger.LogError(tEx.ToString());
					tWireDefinition.Source.Step = "Unknown";
					tWireDefinition.Source.Pin = "Unknown";
				}
				else
				{
					tWireDefinition.Source.Step = tSourceParts[0].Trim();
					tWireDefinition.Source.Pin = tSourceParts[1].Trim();
				}
				string[] tTargetParts = tLineParts[1].Split(new string[] { "." }, StringSplitOptions.None);
				if (tTargetParts.Length == 1)
				{
					FDLException tEx = new FDLException("Syntax error in wire target!", lineNumber + 1, app.FlowText[lineNumber]);
					app.ParseExceptions.Add(tEx);
					app.ErrorLogger.LogError(tEx.ToString());
					tWireDefinition.Target.Step = "Unknown";
					tWireDefinition.Target.Pin = "Unknown";
				}
				else
				{
					tWireDefinition.Target.Step = tTargetParts[0].Trim();
					tWireDefinition.Target.Pin = tTargetParts[1].Replace("{", string.Empty).Trim();
				}

			}
			
			return tWireDefinition;
		}
	}
}
