﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

namespace AB3.Patterns
{
	/// <summary>Options of the asynchron thread
	/// </summary>
	public enum ThreadOptions
	{
		/// <summary>The asynchron thread is a background thread
		/// </summary>
		IsBackgroundThread,

		/// <summary>The asynchron thread is a indipendant thread
		/// </summary>
		IsIndipendantThread
	}

	/// <summary>Options of the locking
	/// </summary>
	public enum LockOptions
	{
		/// <summary>The asynchron thread is locked until end
		/// </summary>
		LockedThread,

		/// <summary>The asynchron thread isn't locked. Create immediately a new.
		/// </summary>
		UnlockedThread
	}

	#region Single arg <T>
	/// <summary>Type safe implementation of a asynchronizer that can be locked to prevent duplicate start
	/// </summary>
	/// <typeparam name="T">Type of the signal</typeparam>
	public class Asynchronizer<T>
	{
		private object _Lock = new object();
		private ThreadOptions _ThreadOption;
		private LockOptions _LockOption;

		/// <summary>Creates a new instance
		/// </summary>
		/// <param name="threadOption">Background or Indipendant</param>
		/// <param name="lockOption">Locked or Unlocked</param>
		public Asynchronizer(ThreadOptions threadOption, LockOptions lockOption)
		{
			_ThreadOption = threadOption;
			_LockOption = lockOption;
		}

		/// <summary>Starts the thread
		/// </summary>
		/// <param name="input">Input to send to the thread</param>
		public void In_Input(T input)
		{
			var tThread = new Thread(() => Process(input));
			if (_ThreadOption == ThreadOptions.IsBackgroundThread)
				tThread.IsBackground = true;

			tThread.Start();
		}

		private void Process(T input)
		{
			if (_LockOption == LockOptions.UnlockedThread)
				Out_Output(input);
			else if (_LockOption == LockOptions.LockedThread)
			{
				lock (this._Lock)
				{
					Out_Output(input);
				}
			}
		}

		/// <summary>Sends the input
		/// </summary>
		public event Action<T> Out_Output;
	}
	#endregion

	#region Trigger ()
	/// <summary>Type safe implementation of a asynchronizer that can be locked to prevent duplicate start
	/// </summary>
	public class Asynchronizer
	{
		private object _Lock = new object();
		private ThreadOptions _ThreadOption;
		private LockOptions _LockOption;

		/// <summary>Creates a new instance
		/// </summary>
		/// <param name="threadOption">Background or Indipendant</param>
		/// <param name="lockOption">Locked or Unlocked</param>
		public Asynchronizer(ThreadOptions threadOption, LockOptions lockOption)
		{
			_ThreadOption = threadOption;
			_LockOption = lockOption;
		}

		/// <summary>Starts the thread
		/// </summary>
		public void In_Input()
		{
			var tThread = new Thread(() => Process());
			if (_ThreadOption == ThreadOptions.IsBackgroundThread)
				tThread.IsBackground = true;

			tThread.Start();
		}

		private void Process()
		{
			if (_LockOption == LockOptions.UnlockedThread)
				Out_Output();
			else if (_LockOption == LockOptions.LockedThread)
			{
				lock (this._Lock)
				{
					Out_Output();
				}
			}
		}

		/// <summary>Sends the signal
		/// </summary>
		public event Action Out_Output;
	}
	#endregion
}
