﻿using System;
using System.Collections.Generic;
using System.Reflection;
using AB3.DataTypes;

namespace AB3.Steps
{
	/// <summary>Activates the instances of steps (not used for static or prepared steps)
	/// </summary>
	public class StepInstancesActivator
	{
		/// <summary>Sends the app definition to the next step
		/// </summary>
		public event Action<AppDefinition> Out_Output;

		/// <summary>Starts activating instances
		/// </summary>
		/// <param name="app">The app definition</param>
		public void In_Input(AppDefinition app)
		{
			#region Logging & Profiling
			app.ProgressLogger.LogInfo("---------------------------------------------------");
			app.ProgressLogger.LogInfo("Start activating step instance.");
			app.ProgressLogger.LogInfo("---------------------------------------------------");
			DateTime tStart = DateTime.Now;
			#endregion

			FDLFlowDefinition tMainFlow = app.FlowDefinitions["MainFlow"];
			ActiveStepInstances(app, "MainFlow", tMainFlow);

			#region Logging & Profiling
			TimeSpan tDuration = DateTime.Now - tStart;
			app.BasicLogger.LogInfo("Finished activating step instances in {0} ms.",
				tDuration.Milliseconds.ToString());
			#endregion

			Out_Output(app);
		}

		#region Helpers: ActivateStepInstances, GetTypeByFDLStepDefinition
		private void ActiveStepInstances(AppDefinition app, string currentNamespace, FDLFlowDefinition currentFlowDefinition)
		{
			app.ProgressLogger.LogInfo("Start building step instances of flow '{0}'", currentFlowDefinition.Name);

			foreach (FDLStepDefinition tStepDefinition in currentFlowDefinition.Steps.Collection.Values)
			{
				if (tStepDefinition.IsSubFlowStep)
				{
					FDLFlowDefinition tSubFlowDefinition = app.FlowDefinitions[tStepDefinition.TypeName];
					string tSubFlowNamespace = string.Concat(currentNamespace, ".", tStepDefinition.CustomName);
					ActiveStepInstances(app, tSubFlowNamespace, tSubFlowDefinition);
				}
				else
				{
					StepInstance tStepInstance = new StepInstance(currentNamespace, tStepDefinition.CustomName);
					if (app.PreparedInstances.ContainsKey(tStepInstance.FullName))
					{
						tStepInstance = app.PreparedInstances[tStepInstance.FullName];
						app.ProgressLogger.LogInfo("Type '{0}' is a prepared instance. Instance taken.", tStepInstance.StepType.ToString());
					}
					else
					{
						tStepInstance.StepType = GetTypeByFDLStepDefinition(app, tStepDefinition);

						if (tStepInstance.StepType.IsAbstract && tStepInstance.StepType.IsSealed) // = IsStatic
						{
							tStepInstance.StepObject = null;
							app.ProgressLogger.LogInfo("Type '{0}' ist static. Don't create instance.", tStepInstance.StepType.ToString());
						}
						else
						{
							try
							{
								app.ProgressLogger.LogInfo("Activate instance of type '{0}'", tStepInstance.StepType.ToString());
								tStepInstance.StepObject = Activator.CreateInstance(tStepInstance.StepType);
							}
							catch (Exception ex)
							{
								app.ErrorLogger.LogError("Activater failed at type '{0}' because of: {1}",
									tStepInstance.StepType.ToString(), ex.Message);
								throw ex;
							}
						}
					}
					app.StepInstances.Add(tStepInstance.FullName, tStepInstance);
				}
			}
		}

		/// <summary>Helper function to find a type given by its step definition
		/// </summary>
		/// <param name="app">App definition</param>
		/// <param name="stepDefinition">Step definition of the type</param>
		/// <returns>Returns the type instance</returns>
		private Type GetTypeByFDLStepDefinition(AppDefinition app, FDLStepDefinition stepDefinition)
		{
			List<Type> tPossibleTypes = new List<Type>();
			string tTypeName;
			if (stepDefinition.TypeName.Contains("."))
				tTypeName = stepDefinition.TypeName; // It's a full qualified name.
			else
				tTypeName = "." + stepDefinition.TypeName; // The . is important to find the type (Namespace.Type)

			foreach (Assembly tAssembly in AppDomain.CurrentDomain.GetAssemblies())
			{
				// 2013-02-28: It's not allowed to examine dynamic assemblies => ignore
				// see: http://stackoverflow.com/questions/14122994/castle-windsor-dynamic-assembly-bug
				if (!tAssembly.IsDynamic)
				{
					foreach (Type tType in tAssembly.GetExportedTypes())
					{
						if (tType.ToString().EndsWith(tTypeName))
							tPossibleTypes.Add(tType);
					}
				}
			}

			if (tPossibleTypes.Count == 1)
				return tPossibleTypes[0];
			else if (tPossibleTypes.Count == 0)
			{
				string tMessage = string.Format("Can't find type '{0}' defined in line {1} in the loaded assemblies!",
					stepDefinition.TypeName, (stepDefinition.LineNumber + 1).ToString());
				app.ErrorLogger.LogError(tMessage);
				throw new ArgumentException(stepDefinition.TypeName, tMessage);
			}
			else
			{
				string tMessage = string.Format("The type name '{0}' is'nt unique! Possible types: ", stepDefinition.TypeName);
				app.ErrorLogger.LogError(tMessage);
				foreach (Type tType in tPossibleTypes)
				{
					app.ErrorLogger.LogError(tType.ToString());
					tMessage += string.Format("[{0}]", tType.ToString());
				}

				throw new ArgumentOutOfRangeException(tMessage);
			}
		}
		#endregion
	}
}
