﻿using System;
using System.Collections.Generic;
using AB3.DataTypes;
using AB3.Logging;

namespace AB3.Steps
{
	/// <summary>Finds the connections of the steps defined in the FDL
	/// </summary>
	public class StepInstancesConnectionsFinder
	{
		/// <summary>Starts the finding of connections
		/// </summary>
		/// <param name="app">The app definition to work with</param>
		public void In_Input(AppDefinition app)
		{
			#region Logging & Profiling
			app.ProgressLogger.LogInfo("---------------------------------------------------");
			app.ProgressLogger.LogInfo("Start finding connections of steps.");
			app.ProgressLogger.LogInfo("---------------------------------------------------");
			DateTime tStart = DateTime.Now;
			#endregion

			FDLFlowDefinition tMainFlow = app.FlowDefinitions["MainFlow"];
			FindConnections(app, "MainFlow", tMainFlow);

			#region Logging & Profiling
			TimeSpan tDuration = DateTime.Now - tStart;
			app.BasicLogger.LogInfo("Finished finding connections of steps in {0} ms.",
				tDuration.Milliseconds.ToString());
			#endregion

			Out_Output(app);
		}

		/// <summary>Sends the app definition to the next step
		/// </summary>
		public event Action<AppDefinition> Out_Output;

		#region Helpers: FindConnections, GetSourceEndPoints, GetTargetEndPoints
		private void FindConnections(AppDefinition app, string fdlNamespace, FDLFlowDefinition fdlFlowDefinition)
		{
			app.ProgressLogger.LogInfo("Start finding connections of flow '{0}'.", fdlFlowDefinition.Name);

			#region Connections of fdlFlowDefinition
			foreach (FDLWireDefinition tConnectionDef in fdlFlowDefinition.Wires.Collection)
			{
				List<ConnectionEndPoint> tSourceEndPoints = new List<ConnectionEndPoint>();
				List<ConnectionEndPoint> tTargetEndPoints = new List<ConnectionEndPoint>();

				if ((tConnectionDef.Source.Step != "In") && (tConnectionDef.Target.Step != "Out"))
				{
					app.ProgressLogger.LogInfo("Start finding source endpoints of connection '{0}':", tConnectionDef.ToString());
					tSourceEndPoints = GetSourceEndPoints(app, fdlNamespace, fdlFlowDefinition, tConnectionDef);

					app.ProgressLogger.LogInfo("Start finding target endpoints of connection '{0}':", tConnectionDef.ToString());
					tTargetEndPoints = GetTargetEndPoints(app, fdlNamespace, fdlFlowDefinition, tConnectionDef);
				}

				#region Store NEW StepInstanceConnections to app.StepInstanceConnections
				StepInstanceConnection tConnection;
				foreach (ConnectionEndPoint tSourceEndPoint in tSourceEndPoints)
				{
					foreach (ConnectionEndPoint tTargetEndPoint in tTargetEndPoints)
					{
						tConnection = new StepInstanceConnection(tSourceEndPoint, tTargetEndPoint);
						if (!app.StepInstanceConnections.ContainsKey(tConnection.FullName))
						{
							app.ProgressLogger.LogInfo("New connection found: '{0}'", tConnection.FullName);
							app.StepInstanceConnections.Add(tConnection.FullName, tConnection);
						}
					}
				}
				#endregion
			}
			#endregion

			#region Connection of subFlowDefinitions
			foreach (FDLStepDefinition tStepDefinition in fdlFlowDefinition.Steps.Collection.Values)
			{
				if (tStepDefinition.IsSubFlowStep)
				{
					string tSubFlowNamespcase = string.Concat(fdlNamespace, ".", tStepDefinition.CustomName);
					FDLFlowDefinition tSubFlowDefinition = app.FlowDefinitions[tStepDefinition.TypeName];
					FindConnections(app, tSubFlowNamespcase, tSubFlowDefinition);
				}
			}
			#endregion
		}

		private List<ConnectionEndPoint> GetSourceEndPoints(AppDefinition app, string currentNamespace, FDLFlowDefinition currentFlowDefinition, FDLWireDefinition currentConnectionDefinition)
		{
			List<ConnectionEndPoint> tSourceEndPoints = new List<ConnectionEndPoint>();

			if (!currentFlowDefinition.Steps.Collection[currentConnectionDefinition.Source.Step].IsSubFlowStep)
			{
				ConnectionEndPoint tConnectionEndPoint = new ConnectionEndPoint();
				tConnectionEndPoint.FDLNamespace = currentNamespace;
				tConnectionEndPoint.FDLStepName = currentConnectionDefinition.Source.Step;
				tConnectionEndPoint.FDLPinName = currentConnectionDefinition.Source.Pin;
				tSourceEndPoints.Add(tConnectionEndPoint);
				app.ProgressLogger.LogInfo("Source endpoint found: {0}", tConnectionEndPoint.FullName);
			}
			else
			{
				string tSubFlowNamespace = string.Concat(currentNamespace, ".", currentConnectionDefinition.Source.Step);
				FDLStepDefinition tSubFlowStepDefinition = currentFlowDefinition.Steps.Collection[currentConnectionDefinition.Source.Step];
				FDLFlowDefinition tSubFlowDefinition = app.FlowDefinitions[tSubFlowStepDefinition.TypeName];

				foreach (FDLWireDefinition tConnection in tSubFlowDefinition.Wires.Collection)
				{
					if ((tConnection.Target.Step == "Out") && (tConnection.Target.Pin == currentConnectionDefinition.Source.Pin))
						tSourceEndPoints.AddRange(GetSourceEndPoints(app, tSubFlowNamespace, tSubFlowDefinition, tConnection));
				}
			}

			return tSourceEndPoints;
		}

		private List<ConnectionEndPoint> GetTargetEndPoints(AppDefinition app, string currentNamespace, FDLFlowDefinition currentFlowDefinition, FDLWireDefinition currentConnectionDefinition)
		{
			List<ConnectionEndPoint> tTargetEndPoints = new List<ConnectionEndPoint>();

			if (!currentFlowDefinition.Steps.Collection[currentConnectionDefinition.Target.Step].IsSubFlowStep)
			{
				ConnectionEndPoint tConnectionEndPoint = new ConnectionEndPoint();
				tConnectionEndPoint.FDLNamespace = currentNamespace;
				tConnectionEndPoint.FDLStepName = currentConnectionDefinition.Target.Step;
				tConnectionEndPoint.FDLPinName = currentConnectionDefinition.Target.Pin;
				tTargetEndPoints.Add(tConnectionEndPoint);
				app.ProgressLogger.LogInfo("Target endpoint found: {0}", tConnectionEndPoint.FullName);
			}
			else
			{
				string tSubFlowNamespace = string.Concat(currentNamespace, ".", currentConnectionDefinition.Target.Step);
				FDLStepDefinition tSubFlowStepDefinition = currentFlowDefinition.Steps.Collection[currentConnectionDefinition.Target.Step];
				FDLFlowDefinition tSubFlowDefinition = app.FlowDefinitions[tSubFlowStepDefinition.TypeName];

				foreach (FDLWireDefinition tConnection in tSubFlowDefinition.Wires.Collection)
				{
					if ((tConnection.Source.Step == "In") && (tConnection.Source.Pin == currentConnectionDefinition.Target.Pin))
						tTargetEndPoints.AddRange(GetTargetEndPoints(app, tSubFlowNamespace, tSubFlowDefinition, tConnection));
				}
			}

			return tTargetEndPoints;
		}
		#endregion
	}
}
