﻿using System;
using System.Reflection;
using AB3.DataTypes;
using AB3.Logging;

namespace AB3.Steps
{
	/// <summary>Connects the source and target steps
	/// </summary>
	public class StepInstancesConnector
	{
		/// <summary>Starts the connecting
		/// </summary>
		/// <param name="app">The app definition to work with</param>
		public void In_Input(AppDefinition app)
		{
			#region Logging & Profiling
			app.ProgressLogger.LogInfo("---------------------------------------------------");
			app.ProgressLogger.LogInfo("Start connecting steps.");
			app.ProgressLogger.LogInfo("---------------------------------------------------");
			DateTime tStart = DateTime.Now;
			#endregion

			StepInstance tSourceStepInstance;
			StepInstance tTartgetStepInstance;
			EventInfo tSourceEventInfo;
			MethodInfo tTargetMethodInfo;
			foreach (StepInstanceConnection tConnection in app.StepInstanceConnections.Values)
			{
				app.ProgressLogger.LogInfo("Connecting '{0}'", tConnection.FullName);
				tSourceStepInstance = app.StepInstances[tConnection.SourceEndPoint.FullStepName];
				tTartgetStepInstance = app.StepInstances[tConnection.TargetEndPoint.FullStepName];

				tSourceEventInfo = tSourceStepInstance.StepType.GetEvent(tConnection.SourceEndPoint.FDLPinName);
				tTargetMethodInfo = tTartgetStepInstance.StepType.GetMethod(tConnection.TargetEndPoint.FDLPinName);

				#region Logging & Throwing Exception
				if (tSourceEventInfo == null)
				{
					string tMessage = string.Format("Can't find source event! Source should be {0}.{1}",
						tSourceStepInstance.StepType.ToString(), tConnection.SourceEndPoint.FDLPinName);
					app.ErrorLogger.LogError(tMessage);
					throw new ArgumentOutOfRangeException("PinName", tMessage);
				}
				else if (tTargetMethodInfo == null)
				{
					string tMessage = string.Format("Can't find target method! Target should be {0}.{1}",
						tTartgetStepInstance.StepType.ToString(), tConnection.TargetEndPoint.FDLPinName);
					app.ErrorLogger.LogError(tMessage);
					throw new ArgumentOutOfRangeException("PinName", tMessage);
				}
				#endregion

				if ((!tSourceEventInfo.GetAddMethod().IsStatic) && (!tTargetMethodInfo.IsStatic))
				{
					#region Establish connection between INSTANCE source event & target INSTANCE method
					Delegate tHandlerDelegate = Delegate.CreateDelegate(tSourceEventInfo.EventHandlerType, tTartgetStepInstance.StepObject, tTargetMethodInfo);
					tSourceEventInfo.AddEventHandler(tSourceStepInstance.StepObject, tHandlerDelegate);
					#endregion
				}
				else if ((tSourceEventInfo.GetAddMethod().IsStatic) && (!tTargetMethodInfo.IsStatic))
				{
					#region Establish connection between STATIC source event & target INSTANCE method
					Delegate tHandlerDelegate = Delegate.CreateDelegate(tSourceEventInfo.EventHandlerType, tTartgetStepInstance.StepObject, tTargetMethodInfo);
					tSourceEventInfo.AddEventHandler(null, tHandlerDelegate);
					#endregion
				}
				else if ((!tSourceEventInfo.GetAddMethod().IsStatic) && (tTargetMethodInfo.IsStatic))
				{
					#region Establish connection between INSTANCE source event & STATIC target method
					Delegate tHandlerDelegate = Delegate.CreateDelegate(tSourceEventInfo.EventHandlerType, null, tTargetMethodInfo);
					tSourceEventInfo.AddEventHandler(tSourceStepInstance.StepObject, tHandlerDelegate);
					#endregion
				}
				else if ((tSourceEventInfo.GetAddMethod().IsStatic) && (tTargetMethodInfo.IsStatic))
				{
					#region Establish connection between STATIC source event & STATIC target method
					Delegate tHandlerDelegate = Delegate.CreateDelegate(tSourceEventInfo.EventHandlerType, null, tTargetMethodInfo);
					tSourceEventInfo.AddEventHandler(null, tHandlerDelegate);
					#endregion
				}
				else
				{
					#region Logging & throw Exception
					string tMessage = "StepInstancesConnector.In_Input: This combination of instance & static is not possible!";
					app.ErrorLogger.LogError(tMessage);
					throw new ArgumentException(tMessage);
					#endregion
				}
			}

			#region Logging & Profiling
			TimeSpan tDuration = DateTime.Now - tStart;
			app.BasicLogger.LogInfo("Finished connecting steps in {0} ms.",
				tDuration.Milliseconds.ToString());
			#endregion

			Out_Output(app);
		}

		/// <summary>Sends the app definition to the next step
		/// </summary>
		public event Action<AppDefinition> Out_Output;
	}
}
