﻿using System.Collections.Generic;
using System.Windows;
using System.Windows.Media;

namespace AB3.AppDesigner.Helpers
{
	/// <summary>
	/// Helper class for VisualTree
	/// <seealso cref="http://stackoverflow.com/questions/4905437/how-can-i-know-if-the-mouse-is-over-the-header-row-in-a-wpf-listview"/>
	/// </summary>
	public static class VisualTreeHelperEx
	{
		/// <summary>
		/// Gets visual parent
		/// </summary>
		/// <typeparam name="T">T</typeparam>
		/// <param name="childObject">The specified Visual child</param>
		/// <returns>Parent of the type T</returns>
		public static T GetVisualParent<T>(object childObject) where T : Visual
		{
			DependencyObject child = childObject as DependencyObject;
			// iteratively traverse the visual tree
			while ((child != null) && !(child is T))
			{
				child = VisualTreeHelper.GetParent(child);
			}
			return child as T;
		}

		/// <summary>
		/// Gets visual child
		/// </summary>
		/// <typeparam name="T">T</typeparam>
		/// <param name="parent">The specified Visual parent</param>
		/// <returns></returns>
		public static T GetVisualChild<T>(object parent) where T : Visual
		{
			DependencyObject dependencyObject = parent as DependencyObject;
			return InternalGetVisualChild<T>(dependencyObject);
		}

		/// <summary>
		/// Gets visual child
		/// </summary>
		/// <typeparam name="T">T</typeparam>
		/// <param name="parent">The specified Visual parent</param>
		/// <returns></returns>
		private static T InternalGetVisualChild<T>(DependencyObject parent) where T : Visual
		{
			T child = default(T);

			int numVisuals = VisualTreeHelper.GetChildrenCount(parent);
			for (int i = 0; i < numVisuals; i++)
			{
				Visual v = (Visual)VisualTreeHelper.GetChild(parent, i);
				child = v as T;
				if (child == null)
				{
					child = GetVisualChild<T>(v);
				}
				if (child != null)
				{
					break;
				}
			}
			return child;
		}


		/// <summary>
		/// Gets all Visual child objects of Type T.
		/// </summary>
		/// <typeparam name="T">T</typeparam>
		/// <param name="parent">The specified Visual parent</param>
		/// <returns>A List containing all Visual childs of type T from parent.</returns>
		public static List<T> GetVisualChildCollection<T>(object parent) where T : Visual
		{
			List<T> visualCollection = new List<T>();
			GetVisualChildCollection(parent as DependencyObject, visualCollection);
			return visualCollection;
		}

		/// <summary>
		/// Gets all Visual child objects of Type T.
		/// </summary>
		/// <typeparam name="T">T</typeparam>
		/// <param name="parent">The specified Visual parent</param>
		/// <param name="visualCollection">A List containing all Visual childs of type T from parent.</param>
		private static void GetVisualChildCollection<T>(DependencyObject parent, List<T> visualCollection) where T : Visual
		{
			int count = VisualTreeHelper.GetChildrenCount(parent);
			for (int i = 0; i < count; i++)
			{
				DependencyObject child = VisualTreeHelper.GetChild(parent, i);
				if (child is T)
				{
					visualCollection.Add(child as T);
				}
				else if (child != null)
				{
					GetVisualChildCollection(child, visualCollection);
				}
			}
		}
	}
}
