﻿using System;
using System.Reflection;
using AB3.DataTypes;
using AB3.Flows;
using AB3.Logging;
using AB3.Steps;

namespace AB3
{
	/// <summary>The main application builder. The entry point of the developer
	/// </summary>
	public class AppBuilder
	{
		#region Fields: App
		/// <summary>Contains the structure, instance of the app to build
		/// </summary>
		public DataTypes.AppDefinition _AppDefinition;
		#endregion

		#region ctors & Initialize
		/// <summary>Creates a new AppBuilder instance without logging
		/// </summary>
		public AppBuilder()
		{
			this._AppDefinition = new DataTypes.AppDefinition();
		}

		/// <summary>Creates a new AppBuilder instance with logging to simple text file
		/// </summary>
		/// <param name="logFileName">Name of the logfile with optional path</param>
		/// <param name="logLevel">Logging level: 0 = Only Error logging, 1 = Basic logging, 2 = Full logging</param>
		public AppBuilder(string logFileName, int logLevel)
		{
			if (logFileName.Contains(@"\")) // Path and name was given
			{
				this.Initialize(new Logging.LogToFile(logFileName), logLevel);
			}
			else // only name of logfile without path was given
			{
				string tAppBuilderLocation = Assembly.GetAssembly(typeof(AppBuilder)).Location;
				string tAppBuilderDirectory = System.IO.Path.GetDirectoryName(tAppBuilderLocation);
				string tLogFileLocation = System.IO.Path.Combine(tAppBuilderDirectory, logFileName);
				this.Initialize(new Logging.LogToFile(logFileName), logLevel);
			}
		}

		/// <summary>Creates a new AppBuilder instance with logging to a custom message logger
		/// </summary>
		/// <param name="logger">Instance of Logging.ILogger to log the progress of the AppBuilder</param>
		/// <param name="logLevel">0 = Only Error logging, 1 = Basic logging, 2 = Full logging</param>
		public AppBuilder(ILogger logger, int logLevel)
		{
			Initialize(logger, logLevel);
		}

		private void Initialize(ILogger logger, int logLevel)
		{
			this._AppDefinition = new DataTypes.AppDefinition();
			this._AppDefinition.ErrorLogger = logger;
			if (logLevel > 0)
				this._AppDefinition.BasicLogger = logger;

			if (logLevel > 1)
				this._AppDefinition.ProgressLogger = logger;
		}
		#endregion

		#region Public methods: LoadFlowsFromString, LoadFlowsFromFile, AddPreparedStepInstance, BuildApp
		/// <summary>Loads a AppBuilder V3 FDL from string
		/// </summary>
		/// <param name="text">Text of the AppBuilder V3 FDL </param>
		public void LoadFlowsFromString(string text)
		{
			FDLTextConverter tTextConverter = new FDLTextConverter();
			FDLParseFlow tParseFlow = new FDLParseFlow();

			tTextConverter.Out_Output += tParseFlow.In_Input;

			tTextConverter.In_Input(_AppDefinition, text);
		}

		/// <summary>Loads a *.ab3 (App Builder V3) file
		/// </summary>
		/// <param name="fileName">filename (with optional path) of the ab3 file</param>
		public void LoadFlowsFromFile(string fileName)
		{
			string tPathFileName;
			if (fileName.Contains(@"\"))
				tPathFileName = fileName;
			else
			{
				string tAppBuilderLocation = Assembly.GetAssembly(typeof(AppBuilder)).Location;
				string tAppBuilderDirectory = System.IO.Path.GetDirectoryName(tAppBuilderLocation);
				tPathFileName = System.IO.Path.Combine(tAppBuilderDirectory, fileName);
			}

			FDLFileLoader tFileReader = new FDLFileLoader();
			FDLParseFlow tParseFlow = new FDLParseFlow();

			tFileReader.Out_Output += tParseFlow.In_Input;

			tFileReader.In_Input(_AppDefinition, tPathFileName);
		}

		/// <summary>Adds a pregenerated step instance to the current app
		/// </summary>
		/// <param name="fdlNamespace">Namespace defined in FDL of instance</param>
		/// <param name="fdlCustomName">Custom name in FDL of instance</param>
		/// <param name="instance">The instance to add</param>
		public void AddPreparedStepInstance(string fdlNamespace, string fdlCustomName, object instance)
		{
			StepInstance tStepInstance = new StepInstance(fdlNamespace, fdlCustomName);
			tStepInstance.StepObject = instance;
			tStepInstance.StepType = instance.GetType();
			_AppDefinition.PreparedInstances.Add(tStepInstance.FullName, tStepInstance);
		}

		/// <summary>Adds a pregenerated step instance to the current app
		/// </summary>
		/// <param name="fdlNamespace">Namespace defined in FDL</param>
		/// <param name="instance">The instance to add</param>
		public void AddPreparedStepInstance(string fdlNamespace, object instance)
		{
			string tTypeName = instance.GetType().ToString();
			string[] tTypeParts = tTypeName.Split(new string[] { "." }, StringSplitOptions.None);
			string tCustomName = tTypeParts[tTypeParts.Length - 1];
			StepInstance tStepInstance = new StepInstance(fdlNamespace, tCustomName);
			tStepInstance.StepObject = instance;
			tStepInstance.StepType = instance.GetType();
			_AppDefinition.PreparedInstances.Add(tStepInstance.FullName, tStepInstance);
		}

		/// <summary>Adds a pregenerated step instance for ALL flows of the current app (like Singleton)
		/// </summary>
		/// <param name="instance">The instance to add</param>
		public void AddPreparedStepInstance(object instance)
		{
			string tTypeName = instance.GetType().ToString();
			string[] tTypeParts = tTypeName.Split(new string[] { "." }, StringSplitOptions.None);
			string tCustomName = tTypeParts[tTypeParts.Length - 1];
			StepInstance tStepInstance = new StepInstance(string.Empty, tCustomName);
			tStepInstance.StepObject = instance;
			tStepInstance.StepType = instance.GetType();
			_AppDefinition.PreparedInstances.Add(tStepInstance.FullName, tStepInstance);
		}

		/// <summary>Builds the app in multiple steps
		/// </summary>
		/// <param name="configureArg">Args to send to all steps that implement IConfigure</param>
		/// <param name="initializeArg">Args to send to all steps that implement IInitialize</param>
		public void BuildApp(object configureArg, object initializeArg)
		{
			_AppDefinition.ConfigureArg = configureArg;
			_AppDefinition.InitializeArg = initializeArg;

			AssembliesFinder tAssembliesFinder = new AssembliesFinder();
			AssembliesLoader tAssembliesLoader = new AssembliesLoader();
			StepInstancesActivator tActivator = new StepInstancesActivator();
			StepInstancesConnectionsFinder tConnectionFinder = new StepInstancesConnectionsFinder();
			StepInstancesConnector tConnector = new StepInstancesConnector();
			StepInstancesConfigurator tConfigurator = new StepInstancesConfigurator();
			StepInstancesInitializer tInitializer = new StepInstancesInitializer();
			StepsEnd tStepsEnd = new StepsEnd("Finished building of the app.");

			tAssembliesFinder.Out_Output += tAssembliesLoader.In_Input;
			tAssembliesLoader.Out_Output += tActivator.In_Input;
			tActivator.Out_Output += tConnectionFinder.In_Input;
			tConnectionFinder.Out_Output += tConnector.In_Input;
			tConnector.Out_Output += tConfigurator.In_Input;
			tConfigurator.Out_Output += tInitializer.In_Input;
			tInitializer.Out_Output += tStepsEnd.In_Input;

			tAssembliesFinder.In_Input(_AppDefinition);
		}
		#endregion
	}
}
