﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace AB3.Patterns
{
	#region JoinToCustomType for 2 signals
	/// <summary>Joins two signals to a single custom type
	/// </summary>
	/// <typeparam name="TInput1">Type of signal 1</typeparam>
	/// <typeparam name="TInput2">Type of signal 2</typeparam>
	/// <typeparam name="TOutput">Type of the output signal</typeparam>
	public class JoinToCustomType<TInput1, TInput2, TOutput>
	{
		private readonly bool autoReset;
		private TInput1 message1;
		private bool message1Present;
		private TInput2 message2;
		private bool message2Present;
		private readonly Func<TInput1, TInput2, TOutput> outputCtor;

		#region Ctor
		/// <summary>Creates a new joiner of two signals to a single custom type with autoreset
		/// </summary>
		public JoinToCustomType(): this(true)
		{ }

		/// <summary>Creates a new joiner of two signals to as single custom type
		/// </summary>
		/// <param name="autoReset">Do you want to autoreset the join</param>
		protected JoinToCustomType(bool autoReset)
		{
			this.autoReset = autoReset;
			var ctor = typeof(TOutput).GetConstructor(new[] { typeof(TInput1), typeof(TInput2) });
			if (ctor == null)
			{
				throw new InvalidOperationException(string.Format(
					"Type {0} needs to have a constructor with two parameters of type {1} and {2} respectively.",
					typeof(TOutput), typeof(TInput1), typeof(TInput2)));
			}
			outputCtor = (x, y) => (TOutput)ctor.Invoke(new object[] { x, y });
		}
		#endregion
		
		/// <summary>Output event
		/// </summary>
		public event Action<TOutput> Out_Output;

		/// <summary>Input pin for signal 1
		/// </summary>
		/// <param name="message">the content of signal 1</param>
		public void In_Input1(TInput1 message)
		{
			message1 = message;
			message1Present = true;
			if (message2Present)
			{
				OutputResult();
			}
		}

		/// <summary>Input pin for signal 2
		/// </summary>
		/// <param name="message">the content of signal 2</param>
		public void In_Input2(TInput2 message)
		{
			message2 = message;
			message2Present = true;
			if (message1Present)
			{
				OutputResult();
			}
		}

		private void OutputResult()
		{
			var result = outputCtor(message1, message2);
			if (autoReset)
			{
				ClearMessagePresence();
			}
			Out_Output(result);
		}

		/// <summary>Clear the current messages
		/// </summary>
		protected void ClearMessagePresence()
		{
			message1Present = false;
			message2Present = false;
		}
	}
	#endregion

	#region Join 3 signals to a single custom type
	/// <summary>Joins three signals to a single custom type
	/// </summary>
	/// <typeparam name="TInput1">Type of signal 1</typeparam>
	/// <typeparam name="TInput2">Type of signal 2</typeparam>
	/// <typeparam name="TInput3">Type of signal 3</typeparam>
	/// <typeparam name="TOutput">Type of the output signal</typeparam>
	public class JoinToCustomType<TInput1, TInput2, TInput3, TOutput>
	{
		private readonly bool autoReset;
		private TInput1 message1;
		private bool message1Present;
		private TInput2 message2;
		private bool message2Present;
		private TInput3 message3;
		private bool message3Present;
		private readonly Func<TInput1, TInput2, TInput3, TOutput> outputCtor;

		#region Ctor
		/// <summary>Creates a new joiner of two signals to as single custom type with autoreset
		/// </summary>
		public JoinToCustomType(): this(false)
		{ }

		/// <summary>Creates a new joiner of three signals to a single custom type
		/// </summary>
		/// <param name="autoReset">Do you want to autoreset the join</param>
		protected JoinToCustomType(bool autoReset)
		{
			this.autoReset = autoReset;
			var ctor = typeof(TOutput).GetConstructor(new[] { typeof(TInput1), typeof(TInput2), typeof(TInput3) });
			if (ctor == null)
			{
				throw new InvalidOperationException(string.Format(
					"Type {0} needs to have a constructor with three parameters of type {1}, {2} and {3} respectively.",
					typeof(TOutput), typeof(TInput1), typeof(TInput2), typeof(TInput3)));
			}
			outputCtor = (x, y, z) => (TOutput)ctor.Invoke(new object[] { x, y, z });
		}
		#endregion

		/// <summary>Output event
		/// </summary>
		public event Action<TOutput> Out_Output;

		/// <summary>Input pin for signal 1
		/// </summary>
		/// <param name="message">the content of signal 1</param>
		public void In_Input1(TInput1 message)
		{
			message1 = message;
			message1Present = true;
			if (message2Present && message3Present)
			{
				OutputResult();
			}
		}

		/// <summary>Input pin for signal 2
		/// </summary>
		/// <param name="message">the content of signal 2</param>
		public void In_Input2(TInput2 message)
		{
			message2 = message;
			message2Present = true;
			if (message1Present && message3Present)
			{
				OutputResult();
			}
		}

		/// <summary>Input pin for signal 3
		/// </summary>
		/// <param name="message">the content of signal 3</param>
		public void In_Input3(TInput3 message)
		{
			message3 = message;
			message3Present = true;
			if (message1Present && message2Present)
			{
				OutputResult();
			}
		}

		private void OutputResult()
		{
			var result = outputCtor(message1, message2, message3);
			if (autoReset)
			{
				ClearMessagePresence();
			}
			Out_Output(result);
		}

		/// <summary>Clear the current messsages
		/// 
		/// </summary>
		protected void ClearMessagePresence()
		{
			message1Present = false;
			message2Present = false;
			message3Present = false;
		}
	}
	#endregion

	#region Join 2 signals to a single custom type with autoreset
	/// <summary>Joins two signals to a single custom type with autoreset
	/// </summary>
	/// <typeparam name="TInput1">Type of signal 1</typeparam>
	/// <typeparam name="TInput2">Type of signal 2</typeparam>
	/// <typeparam name="TOutput">Type of the output signal</typeparam>
	public class AutoResetJoinToCustomType<TInput1, TInput2, TOutput> : JoinToCustomType<TInput1, TInput2, TOutput>
	{
		/// <summary>Creates a new joiner of 2 signals with autoreset
		/// </summary>
		public AutoResetJoinToCustomType(): base(true)
		{ }
	}
	#endregion

	#region Join 2 signals to a single custom type with manual reset
	/// <summary>Joins two signals to a single custom type with autoreset
	/// </summary>
	/// <typeparam name="TInput1">Type of signal 1</typeparam>
	/// <typeparam name="TInput2">Type of signal 2</typeparam>
	/// <typeparam name="TOutput">Type of the output signal</typeparam>
	public class ManualResetJoinToCustomType<TInput1, TInput2, TOutput> : JoinToCustomType<TInput1, TInput2, TOutput>
	{
		/// <summary>Resets the signals
		/// </summary>
		/// <typeparam name="T"></typeparam>
		/// <param name="ignore"></param>
		public void Reset<T>(T ignore)
		{
			Reset();
		}

		/// <summary>Resets the signals
		/// </summary>
		public void Reset()
		{
			ClearMessagePresence();
		}
	}
	#endregion
}
