﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Shapes;

namespace AB3.AppDesigner.Adorners
{
	public class LineAdorner: Adorner
	{
		protected VisualCollection _VisualChildren;
		public Thumb _StartThumb, _TargetThumb, _CenterThumb;
		protected Line _AdornedElement;

		public LineAdorner(Line adornedElement): base(adornedElement)
		{
			_VisualChildren = new VisualCollection(this);
			_AdornedElement = adornedElement;

			BuildAdornerCorner(ref _StartThumb, Cursors.Hand);
			BuildAdornerCorner(ref _TargetThumb, Cursors.Hand);
			BuildAdornerCorner(ref _CenterThumb, Cursors.Hand);

			_StartThumb.DragDelta += StartThumb_DragDelta;
			_StartThumb.PreviewMouseUp += Thumb_PreviewMouseUp;
			_StartThumb.MouseDoubleClick += Thumb_MouseDoubleClick;
			_TargetThumb.DragDelta += TargetThumb_DragDelta;
			_TargetThumb.PreviewMouseUp += Thumb_PreviewMouseUp;
			_TargetThumb.MouseDoubleClick += Thumb_MouseDoubleClick;
			_CenterThumb.MouseDoubleClick += Thumb_MouseDoubleClick;

			MenuItem tAddNewPointAttributeMenuItem = new MenuItem();
			tAddNewPointAttributeMenuItem.Header = "Add new point attribute here";
			tAddNewPointAttributeMenuItem.Click += AddNewPointAttributeMenuItem_Click;

			MenuItem tResetWireMenuItem = new MenuItem();
			tResetWireMenuItem.Header = "Reset wire attritute";
			tResetWireMenuItem.Click += (a, b) => ResetWireAttributes();

			ContextMenu tCenterThumbContectMenu = new ContextMenu();
			tCenterThumbContectMenu.Items.Add(tAddNewPointAttributeMenuItem);
			tCenterThumbContectMenu.Items.Add(tResetWireMenuItem);
			_CenterThumb.ContextMenu = tCenterThumbContectMenu;
		}		

		private void Thumb_MouseDoubleClick(object sender, MouseButtonEventArgs e)
		{
			ThumbDoubleClicked();
		}

		#region EventHandlerActions: DragDelta, PreviewMouseUp, AddNewPointAttributeMenuItem_Click
		private void StartThumb_DragDelta(object sender, DragDeltaEventArgs e)
		{
			StartThumbChanged(_AdornedElement, Mouse.GetPosition(this));
		}

		private void TargetThumb_DragDelta(object sender, DragDeltaEventArgs e)
		{
			TargetThumbChanged(_AdornedElement, Mouse.GetPosition(this));
		}

		private void Thumb_PreviewMouseUp(object sender, MouseButtonEventArgs e)
		{
			ChangeFinished();
		}

		private void AddNewPointAttributeMenuItem_Click(object sender, RoutedEventArgs e)
		{
			Point tPoint = new Point();
			tPoint.X = (_AdornedElement.X1 + _AdornedElement.X2) / 2;
			tPoint.Y = (_AdornedElement.Y1 + _AdornedElement.Y2) / 2;
			AddNewPointAttribute(tPoint);
		}
		#endregion
		
		#region Actions: StartThumbChanged, TargetThumbChanged, ChangeFinished, AddNewPointAttribute, CenterThumbDoubleClicked
		public event Action<Line, Point> StartThumbChanged;
		public event Action<Line, Point> TargetThumbChanged;
		public event Action ChangeFinished;
		public event Action<Point> AddNewPointAttribute;
		public event Action ThumbDoubleClicked;
		public event Action ResetWireAttributes;
		#endregion

		#region Helpers: BuildAdornerCorner
		// Helper method to instantiate the corner Thumbs, set the Cursor property, 
		// set some appearance properties, and add the elements to the visual tree.
		private void BuildAdornerCorner(ref Thumb cornerThumb, Cursor customizedCursor)
		{
			if (cornerThumb != null) return;

			cornerThumb = new Thumb();

			// Set some arbitrary visual characteristics.
			cornerThumb.Cursor = customizedCursor;
			cornerThumb.Height = cornerThumb.Width = 10;
			cornerThumb.Opacity = 0.40;
			cornerThumb.BorderBrush = new SolidColorBrush(Colors.DarkGray);
			cornerThumb.Background = new SolidColorBrush(Colors.MidnightBlue);

			_VisualChildren.Add(cornerThumb);
		}

		// This method ensures that the Widths and Heights are initialized.  Sizing to content produces
		// Width and Height values of Double.NaN.  Because this Adorner explicitly resizes, the Width and Height
		// need to be set first.  It also sets the maximum size of the adorned element.
		private void EnforceSize(FrameworkElement adornedElement)
		{
			if (adornedElement.Width.Equals(Double.NaN))
				adornedElement.Width = adornedElement.DesiredSize.Width;
			if (adornedElement.Height.Equals(Double.NaN))
				adornedElement.Height = adornedElement.DesiredSize.Height;

			FrameworkElement parent = adornedElement.Parent as FrameworkElement;
			if (parent != null)
			{
				adornedElement.MaxHeight = parent.ActualHeight;
				adornedElement.MaxWidth = parent.ActualWidth;
			}
		}
		#endregion

		#region Overrides: ArrangeOverride
		// Arrange the Adorners.
		protected override Size ArrangeOverride(Size finalSize)
		{
			// tDesiredWidth and tDesiredHeight are the width and height of the element that's being adorned.  
			// These will be used to place the ResizingAdorner at the corners of the adorned element.  
			double tDesiredWidth = AdornedElement.RenderSize.Width; //AdornedElement.DesiredSize.Width;
			double tDesiredHeight = AdornedElement.RenderSize.Height; //AdornedElement.DesiredSize.Height;
			// tAdornerWidth & tAdornerHeight are used for placement as well.
			double tAdornerWidth = this.DesiredSize.Width;
			double tAdornerHeight = this.DesiredSize.Height;

			_StartThumb.Arrange(new Rect(_AdornedElement.X1 - tAdornerWidth / 2, _AdornedElement.Y1 - tAdornerHeight / 2, tAdornerWidth, tAdornerHeight));
			_TargetThumb.Arrange(new Rect(_AdornedElement.X2 - tAdornerWidth / 2, _AdornedElement.Y2 - tAdornerHeight / 2, tAdornerWidth, tAdornerHeight));
			_CenterThumb.Arrange(new Rect((_AdornedElement.X2 + _AdornedElement.X1 - tAdornerWidth) / 2, (_AdornedElement.Y2 + _AdornedElement.Y1 - tAdornerHeight) / 2, tAdornerWidth, tAdornerHeight));

			// Return the final size.
			return finalSize;
		}

		// Override the VisualChildrenCount and GetVisualChild properties to interface with 
		// the adorner's visual collection.
		protected override int VisualChildrenCount { get { return _VisualChildren.Count; } }
		protected override Visual GetVisualChild(int index) { return _VisualChildren[index]; }
		#endregion
	}
}
